using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net;
using System.Threading;

using VMware.Vim;

using FileInfo = System.IO.FileInfo;

namespace Nant.VMWare
{
    public class SphereVirtualMachine : IVirtualMachine
    {
        private readonly GuestOperationsManager _manager;
        private readonly VirtualMachine _virtualMachine;
        private NamePasswordAuthentication _authentication;
        private GuestFileManager _fileManager;

        public SphereVirtualMachine( VirtualMachine virtualMachine )
        {
            if ( virtualMachine == null )
            {
                throw new ArgumentNullException( "virtualMachine", "A valid vSphere virtual machine object must be provided." );
            }

            _virtualMachine = virtualMachine;
            _virtualMachine.UpdateViewData();
            _manager = new GuestOperationsManager( _virtualMachine.Client, _virtualMachine.Client.ServiceContent.GuestOperationsManager );
        }

        #region Implementation of IVirtualMachine

        public void Start()
        {
            if ( !VirtualMachineIsRunning() )
            {
                _virtualMachine.PowerOnVM( null );
            }

            while ( !VirtualMachineIsRunning() )
            {
                Thread.Sleep( 1000 );
            }
        }

        public void Login( string login, string password )
        {
            if ( !VirtualMachineIsRunning() )
            {
                throw new VirtualMachineIsNotReadyException( this, "Virtual machine must be powered on first." );
            }

            _manager.UpdateViewData();

            GuestAuthManager auth = new GuestAuthManager( _virtualMachine.Client, _manager.AuthManager );
            _authentication = new NamePasswordAuthentication
                                  {
                                      Username = login,
                                      Password = password,
                                      InteractiveSession = true
                                  };
            while ( _virtualMachine.Guest.ToolsStatus.GetValueOrDefault( VirtualMachineToolsStatus.toolsNotRunning ) != VirtualMachineToolsStatus.toolsOk
                    || !_virtualMachine.Guest.InteractiveGuestOperationsReady.GetValueOrDefault( false ) )
            {
                Thread.Sleep( 1000 );
                _virtualMachine.UpdateViewData();

                if ( !VirtualMachineIsRunning() )
                {
                    throw new VirtualMachineIsNotReadyException( this, "Virtual machine must be powered on first." );
                }
            }
            auth.ValidateCredentialsInGuest( _virtualMachine.MoRef, _authentication );
            _fileManager = new GuestFileManager( _virtualMachine.Client, _manager.FileManager );
        }

        public void CopyFileToVm( string sourceFilePath, string destinationFilePath )
        {
            if ( !LoggedInToVirtualMachine() )
            {
                throw new VirtualMachineIsNotReadyException( this, "Not logged in." );
            }
            string folderToCopyFileTo = Path.GetDirectoryName( destinationFilePath );
            string folderOfFolderToCopyFileTo = Path.GetDirectoryName( folderToCopyFileTo );
            string folderNameToCopyFileTo = Path.GetFileName( folderToCopyFileTo );

            bool exists = false;

            if ( !string.IsNullOrEmpty( folderOfFolderToCopyFileTo ) )
            {
                try
                {
                    GuestListFileInfo info = _fileManager.ListFilesInGuest(
                        _virtualMachine.MoRef,
                        _authentication,
                        folderOfFolderToCopyFileTo,
                        null,
                        null,
                        folderNameToCopyFileTo );

                    exists = ( info.Files != null && info.Files.Length > 0 && info.Files.Any( f => f.Path == Path.GetFileName( folderToCopyFileTo ) ) );
                }
                catch ( VimException )
                {
                }
            }
            else
            {
                exists = true;
            }

            try
            {
                if ( !exists )
                {
                    _fileManager.MakeDirectoryInGuest( _virtualMachine.MoRef, _authentication, folderToCopyFileTo, true );
                }
            }
            catch ( VimException )
            {
            }

            string url = _fileManager.InitiateFileTransferToGuest(
                _virtualMachine.MoRef,
                _authentication,
                destinationFilePath,
                new GuestWindowsFileAttributes(),
                new FileInfo( sourceFilePath ).Length,
                true );

            UploadFile( sourceFilePath, url.Replace( "*", new Uri( _virtualMachine.Client.ServiceUrl ).Host ) );
        }


        public void CopyFileFromVm( string sourceFilePath, string destinationFilePath )
        {
            if ( !LoggedInToVirtualMachine() )
            {
                throw new VirtualMachineIsNotReadyException( this, "Not logged in." );
            }

            FileTransferInformation info;

            try
            {
                info = _fileManager.InitiateFileTransferFromGuest( _virtualMachine.MoRef, _authentication, sourceFilePath );
            }
            catch ( VimException ex )
            {
                if ( ex.MethodFault is FileNotFound )
                {
                    throw new FileNotFoundException( "File does not exist in Virtual Machine.", sourceFilePath );
                }
                throw;
            }

            DownloadFile( info.Url.Replace( "*", new Uri( _virtualMachine.Client.ServiceUrl ).Host ), destinationFilePath );
        }

        public string ReadFileContentsFromVm( string filePathInVm )
        {
            string tempFile = Path.GetTempFileName();
            string contents;

            try
            {
                CopyFileFromVm( filePathInVm, tempFile );
                contents = File.ReadAllText( tempFile );
            }
            finally
            {
                File.Delete( tempFile );
            }

            return contents;
        }

        public bool FileExistsInVm( string filePathInVm )
        {
            if ( !LoggedInToVirtualMachine() )
            {
                throw new VirtualMachineIsNotReadyException( this, "Not logged in." );
            }

            GuestListFileInfo info = _fileManager.ListFilesInGuest( _virtualMachine.MoRef,
                                                                    _authentication,
                                                                    Path.GetDirectoryName( filePathInVm ),
                                                                    null, null,
                                                                    Path.GetFileName( filePathInVm ) );

            return info.Files != null && info.Files.Length > 0;
        }

        public void TakeScreenshot( string screenshotLocation )
        {
            if ( !VirtualMachineIsRunning() )
            {
                throw new VirtualMachineIsNotReadyException( this, "Not running." );
            }

            string url = _virtualMachine.CreateScreenshot();

            string requestUrl = string.Format(
                "https://{0}/folder/{1}?dcPath=ha-datacenter&dsName={2}",
                new Uri( _virtualMachine.Client.ServiceUrl ).Host, url.Replace( "[" + _virtualMachine.Config.DatastoreUrl.First().Name + "] ", "" ),
                _virtualMachine.Config.DatastoreUrl.First().Name );

            DownloadFile( requestUrl, screenshotLocation );
        }

        public void Shutdown()
        {
            if ( VirtualMachineIsRunning() )
            {
                _virtualMachine.ShutdownGuest();
            }
        }

        public void RevertToSnapshot( string snapshotName )
        {
            _virtualMachine.UpdateViewData();

            VirtualMachineSnapshot cleanSnapshot = FindSnapshot(
                snapshotName, _virtualMachine.Snapshot.RootSnapshotList );

            if ( cleanSnapshot == null )
            {
                throw new ApplicationException( "Unable to find a snapshot named " + snapshotName );
            }

            _virtualMachine.UpdateViewData();

            cleanSnapshot.RevertToSnapshot( null, false );
        }

        public void TakeSnapshot( string name, string description )
        {
            _virtualMachine.CreateSnapshot( name, description, true, true );
        }

        public void ConnectIso( string isoLocation )
        {
            const string remotePath = "/Temp";

            UploadArtifact( isoLocation, remotePath );

            _virtualMachine.UpdateViewData();

            VirtualCdrom cd = _virtualMachine.Config.Hardware.Device.FirstOrDefault( d => d is VirtualCdrom ) as VirtualCdrom;

            if ( cd == null )
            {
                throw new ApplicationException( "CD is null" );
            }

            VirtualCdromIsoBackingInfo newBacking = new VirtualCdromIsoBackingInfo
                                                        {
                                                            FileName = string.Format( "[{0}] {1}", "datastore1", remotePath )
                                                        };
            cd.Backing = newBacking;
            cd.Connectable.StartConnected = true;
            cd.Connectable.Connected = true;

            VirtualDeviceConfigSpec deviceSpec = new VirtualDeviceConfigSpec
                                                     {
                                                         Device = cd,
                                                         Operation = VirtualDeviceConfigSpecOperation.edit
                                                     };

            VirtualMachineConfigSpec vmSpec = new VirtualMachineConfigSpec
                                                  {
                                                      DeviceChange = new[] {deviceSpec}
                                                  };

            _virtualMachine.ReconfigVM( vmSpec );
        }

        public void StartProcess( string processLocationInVm, string arguments )
        {
            GuestProcessManager processManager = new GuestProcessManager( _virtualMachine.Client, _manager.ProcessManager );

            GuestProgramSpec args = new GuestProgramSpec
                                        {
                                            Arguments = arguments,
                                            ProgramPath = processLocationInVm,
                                            WorkingDirectory = Path.GetDirectoryName( processLocationInVm )
                                        };

            processManager.StartProgramInGuest( _virtualMachine.MoRef, _authentication, args );
        }

        public int StartProcessAndWaitForExit( string processLocationInVm, string arguments )
        {
            GuestProcessManager processManager = new GuestProcessManager( _virtualMachine.Client, _manager.ProcessManager );

            GuestProgramSpec args = new GuestProgramSpec
                                        {
                                            Arguments = arguments,
                                            ProgramPath = processLocationInVm,
                                            WorkingDirectory = Path.GetDirectoryName( processLocationInVm )
                                        };

            long pid = processManager.StartProgramInGuest( _virtualMachine.MoRef, _authentication, args );
            int exitCode = 0;

            while ( true )
            {
                GuestProcessInfo process = processManager.ListProcessesInGuest( _virtualMachine.MoRef, _authentication, new[] {pid} )
                    .FirstOrDefault( p => p.Pid == pid );

                if ( process == null || process.EndTime != null )
                {
                    if ( process != null && process.ExitCode.HasValue )
                    {
                        exitCode = process.ExitCode.Value;
                    }
                    break;
                }

                Thread.Sleep( 6000 );
                processManager.UpdateViewData();
            }

            return exitCode;
        }

        public bool VirtualMachineIsRunning()
        {
            _virtualMachine.UpdateViewData();

            while ( _virtualMachine.Guest.GuestState != "running"
                    && _virtualMachine.Guest.GuestState != "notRunning" )
            {
                Thread.Sleep( 500 );
                _virtualMachine.UpdateViewData();
            }
            return _virtualMachine.Guest.GuestOperationsReady.HasValue && _virtualMachine.Guest.GuestOperationsReady.Value;
        }

        public bool LoggedInToVirtualMachine()
        {
            return _authentication != null && _fileManager != null;
        }

        private VirtualMachineSnapshot FindSnapshot( string name, IEnumerable<VirtualMachineSnapshotTree> trees )
        {
            VirtualMachineSnapshot snapshot = null;

            foreach ( VirtualMachineSnapshotTree tree in trees )
            {
                if ( tree.Name == name )
                {
                    snapshot = new VirtualMachineSnapshot( _virtualMachine.Client, tree.Snapshot );
                    break;
                }

                snapshot = FindSnapshot( name, tree.ChildSnapshotList );

                if ( snapshot != null )
                {
                    break;
                }
            }

            return snapshot;
        }

        private void UploadFile( string from, string to )
        {
            string cookie = _virtualMachine.Client.VimService.CookieContainer.GetCookies( new Uri( _virtualMachine.Client.VimService.Url ) )[ 0 ].ToString();

            WebClient wc = new WebClient
                               {
                                   Credentials = _virtualMachine.Client.VimService.Credentials
                               };
            wc.Headers.Add( HttpRequestHeader.Cookie, cookie );

            while ( true )
            {
                try
                {
                    wc.UploadFile( to, "PUT", from );
                    break;
                }
                catch ( Exception ex )
                {
                    Console.WriteLine( ex.Message );

                    if ( ex.InnerException != null )
                    {
                        Console.WriteLine( ex.InnerException.Message );
                    }

                    Thread.Sleep( 5000 );
                }
            }

            wc.Dispose();
        }

        private void UploadArtifact( string sourceFilePath, string destinationPath )
        {
            string serviceUrl = _virtualMachine.Client.VimService.Url;

            serviceUrl = serviceUrl.Substring( 0, serviceUrl.LastIndexOf( "sdk" ) - 1 );

            string localFilePath = Path.GetFullPath( sourceFilePath );

            String httpUrl = serviceUrl + "/folder/" + destinationPath + "?dcPath=ha-datacenter&dsName=" + "datastore1";

            httpUrl = httpUrl.Replace( " ", " % 20" );

            string cookie = _virtualMachine.Client.VimService.CookieContainer.GetCookies( new Uri( _virtualMachine.Client.VimService.Url ) )[ 0 ].ToString();

            HttpWebRequest request = ( HttpWebRequest )WebRequest.Create( httpUrl );
            request.Method = "PUT";
            request.Credentials = _virtualMachine.Client.VimService.Credentials;
            request.Headers.Add( HttpRequestHeader.Cookie, cookie );
            request.Timeout = Timeout.Infinite;
            request.ReadWriteTimeout = Timeout.Infinite;
            request.ContentType = "application/octet-stream";
            request.AllowWriteStreamBuffering = false;

            const int bytes = 1000;
            byte[] buffer = new byte[bytes];

            using ( FileStream fs = new FileStream( localFilePath, FileMode.Open, FileAccess.Read ) )
            {
                request.ContentLength = fs.Length;

                using ( Stream destination = request.GetRequestStream() )
                {
                    int read;
                    while ( ( read = fs.Read( buffer, 0, bytes ) ) != 0 )
                    {
                        destination.Write( buffer, 0, read );
                    }

                    destination.Close();
                }
                fs.Close();
            }
        }

        private void DownloadFile( string from, string to )
        {
            string cookie = _virtualMachine.Client.VimService.CookieContainer.GetCookies( new Uri( _virtualMachine.Client.VimService.Url ) )[ 0 ].ToString();

            WebClient webCLient = new WebClient
                                      {
                                          Credentials = _virtualMachine.Client.VimService.Credentials
                                      };
            webCLient.Headers.Add( HttpRequestHeader.Cookie, cookie );

            try
            {
                webCLient.DownloadFile( from, to );
            }
            catch ( Exception ex )
            {
                Console.WriteLine( ex.Message );

                if ( ex.InnerException != null )
                {
                    Console.WriteLine( ex.InnerException.Message );
                }
            }
            webCLient.Dispose();
        }

        #endregion
    }
}