﻿using System.IO;

using LibGit2Sharp;

using NAnt.Core;

using NUnit.Framework;

namespace Nant.Git
{
    [TestFixture]
    [Category( TestCategories.CommitStage )]
    public class GetVersionNumberBasedOnTagsTests
    {
        #region Setup/Teardown

        [SetUp]
        public void CreateTsk()
        {
            _task = new TestObject();
        }

        #endregion

        public class TestObject : GetVersionNumberBasedOnTags
        {
            private readonly PropertyDictionary _properties = new PropertyDictionary( null );

            public override PropertyDictionary Properties
            {
                get
                {
                    return _properties;
                }
            }

            public void TestExecute()
            {
                ExecuteTask();
            }
        }

        private TestObject _task;

        private void DeleteRecursiveFolder( string pFolderPath )
        {
            foreach ( string Folder in Directory.GetDirectories( pFolderPath ) )
            {
                DeleteRecursiveFolder( Folder );
            }

            foreach ( string file in Directory.GetFiles( pFolderPath ) )
            {
                FileInfo fi = new FileInfo( file );
                File.SetAttributes( file, FileAttributes.Normal );
                File.Delete( file );
            }

            Directory.Delete( pFolderPath );
        }

        [Test]
        public void ReturnsHashOfLastCommitAndVersionNumber()
        {
            if ( Directory.Exists( "test" ) )
            {
                DeleteRecursiveFolder( "test" );
            }

            Directory.CreateDirectory( "test" );

            Repository repo = Repository.Init( "test" );
            Commit commit = repo.Commit( "haha" );
            repo.ApplyTag( "odd" );

            commit = repo.Commit( "haha1" );
            commit = repo.Commit( "haha2" );

            try
            {
                _task.Properties.Add( GetVersionNumberBasedOnTags.RepositoryNameProperty, "test" );
                _task.Properties.Add( GetVersionNumberBasedOnTags.MajorMinorVersionProperty, "v1.3" );
                _task.TestExecute();
            }
            catch
            {
                throw;
            }
            finally
            {
                repo.Dispose();

                if ( Directory.Exists( "test" ) )
                {
                    DeleteRecursiveFolder( "test" );
                }
            }

            Assert.AreEqual( commit.Sha, _task.Properties[ GetVersionNumberBasedOnTags.RevisionHashProperty ] );
            Assert.AreEqual( "1.3.0.3", _task.Properties[ GetVersionNumberBasedOnTags.VersionNumberProperty ] );
        }

        [Test]
        public void ReturnsHashOfLastCommitAndVersionNumberWhenThereAreLotsOfCommits()
        {
            if ( Directory.Exists( "test" ) )
            {
                DeleteRecursiveFolder( "test" );
            }

            Directory.CreateDirectory( "test" );

            Repository repo = Repository.Init( "test" );
            Commit commit = null;

            for ( int i = 0; i < 1000; i++ )
            {
                commit = repo.Commit( "ahha" + i );

                if ( i == 2 )
                {
                    repo.ApplyTag( "odd" );
                }
            }

            try
            {
                _task.Properties.Add( GetVersionNumberBasedOnTags.RepositoryNameProperty, "test" );
                _task.Properties.Add( GetVersionNumberBasedOnTags.MajorMinorVersionProperty, "v1.3.5.9" );
                _task.TestExecute();
            }
            catch
            {
                throw;
            }
            finally
            {
                repo.Dispose();

                if ( Directory.Exists( "test" ) )
                {
                    DeleteRecursiveFolder( "test" );
                }
            }

            Assert.AreEqual( commit.Sha, _task.Properties[ GetVersionNumberBasedOnTags.RevisionHashProperty ] );
            Assert.AreEqual( "1.3.5." + ( 1000 - 2 ), _task.Properties[ GetVersionNumberBasedOnTags.VersionNumberProperty ] );
        }

        [Test]
        public void ReturnsHashOfLastCommitAndVersionNumberWithZerosByDefault()
        {
            if ( Directory.Exists( "test" ) )
            {
                DeleteRecursiveFolder( "test" );
            }

            Directory.CreateDirectory( "test" );

            Repository repo = Repository.Init( "test" );
            Commit commit = repo.Commit( "haha" );
            repo.ApplyTag( "odd" );

            try
            {
                _task.Properties.Add( GetVersionNumberBasedOnTags.RepositoryNameProperty, "test" );
                _task.Properties.Add( GetVersionNumberBasedOnTags.MajorMinorVersionProperty, "" );
                _task.TestExecute();
            }
            catch
            {
                throw;
            }
            finally
            {
                repo.Dispose();

                if ( Directory.Exists( "test" ) )
                {
                    DeleteRecursiveFolder( "test" );
                }
            }

            Assert.AreEqual( commit.Sha, _task.Properties[ GetVersionNumberBasedOnTags.RevisionHashProperty ] );
            Assert.AreEqual( "0.0.0.1", _task.Properties[ GetVersionNumberBasedOnTags.VersionNumberProperty ] );
        }

        [Test]
        [ExpectedException( typeof( BuildException ) )]
        public void ThrowsAnExceptionWhenBaseVersionPropertyIsNull()
        {
            _task.Properties.Add( GetVersionNumberBasedOnTags.RepositoryNameProperty, "" );
            _task.Properties.Add( GetVersionNumberBasedOnTags.MajorMinorVersionProperty, null );
            _task.TestExecute();
        }


        [Test]
        [ExpectedException( typeof( BuildException ) )]
        public void ThrowsAnExceptionWhenNoBaseVersionPropertyIsSupplied()
        {
            _task.Properties.Add( GetVersionNumberBasedOnTags.RepositoryNameProperty, "someoddrepo" );
            _task.TestExecute();
        }

        [Test]
        [ExpectedException( typeof( BuildException ) )]
        public void ThrowsAnExceptionWhenNoRepositoryPropertyIsSupplied()
        {
            _task.TestExecute();
        }

        [Test]
        [ExpectedException( typeof( BuildException ) )]
        public void ThrowsAnExceptionWhenRepositoryPropertyIsNull()
        {
            _task.Properties.Add( GetVersionNumberBasedOnTags.RepositoryNameProperty, null );
            _task.TestExecute();
        }

        [Test]
        [ExpectedException( typeof( BuildException ) )]
        public void ThrowsAnExceptionWhenRepositoryPropertyPointsToNonGitFolder()
        {
            _task.Properties.Add( GetVersionNumberBasedOnTags.RepositoryNameProperty, "" );
            _task.TestExecute();
        }
    }
}