﻿using System;
using System.Drawing;
using System.IO;
using System.Net;

using NUnit.Framework;

using VMware.Vim;

namespace Nant.VMWare
{
    internal partial class SphereVirtualMachineTests
    {
        private byte[] ReadFileFromVirtualMachine( string fileInVirtualMachine )
        {
            _vmWareVirtualMachine.UpdateViewData();

            GuestOperationsManager operationsInGuestManager = new GuestOperationsManager( _vmWareVirtualMachine.Client,
                                                                                          _vmWareVirtualMachine.Client.ServiceContent.
                                                                                              GuestOperationsManager );
            operationsInGuestManager.UpdateViewData();

            GuestFileManager fileManager = new GuestFileManager( _vmWareVirtualMachine.Client, operationsInGuestManager.FileManager );
            NamePasswordAuthentication authentication = new NamePasswordAuthentication
                                                            {
                                                                Username = VirtualMachineUserName,
                                                                Password = VirtualMachineUserPassword,
                                                                InteractiveSession = false
                                                            };
            string url = fileManager.InitiateFileTransferFromGuest( _vmWareVirtualMachine.MoRef, authentication, fileInVirtualMachine ).Url;
            url = url.Replace( "*", FarmUrl );

            string cookie =
                _vmWareVirtualMachine.Client.VimService.CookieContainer.GetCookies( new Uri( _vmWareVirtualMachine.Client.VimService.Url ) )[ 0 ].ToString();

            WebClient webClient = new WebClient
                                      {
                                          Credentials = _vmWareVirtualMachine.Client.VimService.Credentials
                                      };
            webClient.Headers.Add( HttpRequestHeader.Cookie, cookie );
            byte[] fileData = webClient.DownloadData( url );
            webClient.Dispose();

            return fileData;
        }

        [Test]
        public void CopiesFileFromVirtualMachineWhenCopyFileFromVmIsCalledAndFileExistsInVirtualMachine()
        {
            IVirtualMachine vm = CreateVirtualMachine();
            vm.Start();
            vm.Login( VirtualMachineUserName, VirtualMachineUserPassword );

            string file = Path.GetTempFileName();
            string destinationFile = "c:\\" + Path.GetFileName( file );

            File.WriteAllText( file, "OHOHO" );
            byte[] bytes = File.ReadAllBytes( file );

            try
            {
                vm.CopyFileToVm( file, destinationFile );
            }
            finally
            {
                File.Delete( file );
            }

            file = Path.GetTempFileName();

            try
            {
                vm.CopyFileFromVm( destinationFile, file );

                Assert.AreEqual( bytes, File.ReadAllBytes( file ),
                                 "Downloaded file content differs from the content that was expected." );
            }
            finally
            {
                File.Delete( file );
            }
        }

        [Test]
        public void CopiesFileToVirtualMachineWhenCopyFileToVmIsCalledAndSourceFileExists()
        {
            IVirtualMachine vm = CreateVirtualMachine();
            vm.Start();
            vm.Login( VirtualMachineUserName, VirtualMachineUserPassword );

            string file = Path.GetTempFileName();
            string destinationFile = "c:\\" + Path.GetFileName( file );
            File.WriteAllText( file, "OHOHO" );

            try
            {
                vm.CopyFileToVm( file, destinationFile );

                Assert.AreEqual( File.ReadAllBytes( file ), ReadFileFromVirtualMachine( destinationFile ),
                                 "Downloaded file content differs from the content that was sent" );
            }
            finally
            {
                File.Delete( file );
            }
        }

        [Test]
        public void CreatesDirectoryInVirtualMachineWhenCopyFileToVmIsCalledAndDirectoryToCopyToDoesNotExist()
        {
            IVirtualMachine vm = CreateVirtualMachine();
            vm.Start();
            vm.Login( VirtualMachineUserName, VirtualMachineUserPassword );

            string file = Path.GetTempFileName();
            string destinationFile = "c:\\dir1\\dir2\\dir3\\" + Path.GetFileName( file );
            File.WriteAllText( file, "OHOHO" );

            try
            {
                vm.CopyFileToVm( file, destinationFile );

                Assert.AreEqual( File.ReadAllBytes( file ), ReadFileFromVirtualMachine( destinationFile ),
                                 "Downloaded file content differs from the content that was sent" );
            }
            finally
            {
                File.Delete( file );
            }
        }

        [Test]
        public void GetsFileContentsFromVirtualMachineWhenReadFileContentsFromVmIsCalledAndFileExistsInVirtualMachine()
        {
            IVirtualMachine vm = CreateVirtualMachine();
            vm.Start();
            vm.Login( VirtualMachineUserName, VirtualMachineUserPassword );

            string file = Path.GetTempFileName();
            string destinationFile = "c:\\" + Path.GetFileName( file );

            const string fileContents = "OHOHO";
            File.WriteAllText( file, fileContents );

            try
            {
                vm.CopyFileToVm( file, destinationFile );
            }
            finally
            {
                File.Delete( file );
            }

            Assert.AreEqual( fileContents, vm.ReadFileContentsFromVm( destinationFile ),
                             "Downloaded file content differs from the content that was expected." );
        }

        [Test]
        public void ReturnsFalseWhenCallingFileExistsInVmAndFileDoesNotExist()
        {
            IVirtualMachine vm = CreateVirtualMachine();
            vm.Start();
            vm.Login( VirtualMachineUserName, VirtualMachineUserPassword );

            const string file = "c:\\ohoho.file";
            Assert.IsFalse( vm.FileExistsInVm( file ) );
        }

        [Test]
        public void ReturnsTrueWhenCallingFileExistsInVmAndFileExists()
        {
            IVirtualMachine vm = CreateVirtualMachine();
            vm.Start();
            vm.Login( VirtualMachineUserName, VirtualMachineUserPassword );

            string tempFile = Path.GetTempFileName();
            string file = "c:\\" + Path.GetFileName( tempFile );

            try
            {
                vm.CopyFileToVm( tempFile, file );
                Assert.IsTrue( vm.FileExistsInVm( file ) );
            }
            finally
            {
                File.Delete( tempFile );
            }
        }

        [Test]
        public void SavesScreenshotOfRunningVmWhenTakeScreenshotIsCalled()
        {
            IVirtualMachine vm = CreateVirtualMachine();
            vm.Start();

            string file = Path.GetTempFileName();

            try
            {
                vm.TakeScreenshot( file );
                using ( Image screenshot = Image.FromFile( file ) )
                {
                    Assert.IsNotNull( screenshot );
                    Assert.AreNotEqual( SizeF.Empty, screenshot.PhysicalDimension );
                }
            }
            finally
            {
                File.Delete( file );
            }
        }

        [Test]
        [ExpectedException( typeof( VirtualMachineIsNotReadyException ) )]
        public void ThrowsAnExceptionWhenCallingFileExistsInVmBeforeLogin()
        {
            IVirtualMachine vm = CreateVirtualMachine();
            vm.Start();

            const string file = "c:\\ohoho.file";
            Assert.IsFalse( vm.FileExistsInVm( file ) );
        }

        [Test]
        [ExpectedException( typeof( VirtualMachineIsNotReadyException ) )]
        public void ThrowsAnExceptionWhenCallingFileExistsInVmOnStoppedVirtualMachine()
        {
            IVirtualMachine vm = CreateVirtualMachine();

            const string file = "c:\\ohoho.file";
            Assert.IsFalse( vm.FileExistsInVm( file ) );
        }

        [Test]
        [ExpectedException( typeof( FileNotFoundException ) )]
        public void ThrowsAnExceptionWhenTryingToCopyFileFromVirtualMachineAndFileDoesNotExist()
        {
            IVirtualMachine vm = CreateVirtualMachine();
            vm.Start();
            vm.Login( VirtualMachineUserName, VirtualMachineUserPassword );

            const string file = "c:\\ohoho.file";
            string tempFile = Path.GetTempFileName();
            try
            {
                vm.CopyFileFromVm( file, tempFile );
            }
            finally
            {
                File.Delete( tempFile );
            }
        }

        [Test]
        [ExpectedException( typeof( VirtualMachineIsNotReadyException ) )]
        public void ThrowsAnExceptionWhenTryingToCopyFileFromVirtualMachineBeforeLogin()
        {
            IVirtualMachine vm = CreateVirtualMachine();
            vm.Start();

            const string file = "c:\\ohoho.file";
            vm.CopyFileFromVm( file, "c:\\" + Path.GetFileName( file ) );
        }

        [Test]
        [ExpectedException( typeof( VirtualMachineIsNotReadyException ) )]
        public void ThrowsAnExceptionWhenTryingToCopyFileFromVirtualMachineOnStoppedVirtualMachine()
        {
            IVirtualMachine vm = CreateVirtualMachine();

            const string file = "c:\\ohoho.file";
            vm.CopyFileFromVm( file, "c:\\" + Path.GetFileName( file ) );
        }

        [Test]
        [ExpectedException( typeof( FileNotFoundException ) )]
        public void ThrowsAnExceptionWhenTryingToCopyFileToVirtualMachineAndFileDoesNotExist()
        {
            IVirtualMachine vm = CreateVirtualMachine();
            vm.Start();
            vm.Login( VirtualMachineUserName, VirtualMachineUserPassword );

            const string file = "c:\\ohoho.file";

            vm.CopyFileToVm( file, "c:\\" + Path.GetFileName( file ) );
        }

        [Test]
        [ExpectedException( typeof( VirtualMachineIsNotReadyException ) )]
        public void ThrowsAnExceptionWhenTryingToCopyFileToVirtualMachineBeforeLogin()
        {
            IVirtualMachine vm = CreateVirtualMachine();
            vm.Start();

            string file = Path.GetTempFileName();
            try
            {
                vm.CopyFileToVm( file, "c:\\" + Path.GetFileName( file ) );
            }
            finally
            {
                File.Delete( file );
            }
        }

        [Test]
        [ExpectedException( typeof( VirtualMachineIsNotReadyException ) )]
        public void ThrowsAnExceptionWhenTryingToCopyFileToVirtualMachineOnStoppedVirtualMachine()
        {
            IVirtualMachine vm = CreateVirtualMachine();

            string file = Path.GetTempFileName();
            try
            {
                vm.CopyFileToVm( file, "c:\\" + Path.GetFileName( file ) );
            }
            finally
            {
                File.Delete( file );
            }
        }

        [Test]
        [ExpectedException( typeof( FileNotFoundException ) )]
        public void ThrowsAnExceptionWhenTryingToReadFileContentsFromVmAndFileDoesNotExist()
        {
            IVirtualMachine vm = CreateVirtualMachine();
            vm.Start();
            vm.Login( VirtualMachineUserName, VirtualMachineUserPassword );

            const string file = "c:\\ohoho.file";
            vm.ReadFileContentsFromVm( file );
        }

        [Test]
        [ExpectedException( typeof( VirtualMachineIsNotReadyException ) )]
        public void ThrowsAnExceptionWhenTryingToReadFileContentsFromVmBeforeLogin()
        {
            IVirtualMachine vm = CreateVirtualMachine();
            vm.Start();

            const string file = "c:\\ohoho.file";
            vm.ReadFileContentsFromVm( file );
        }

        [Test]
        [ExpectedException( typeof( VirtualMachineIsNotReadyException ) )]
        public void ThrowsAnExceptionWhenTryingToReadFileContentsFromVmOnStoppedVirtualMachine()
        {
            IVirtualMachine vm = CreateVirtualMachine();

            const string file = "c:\\ohoho.file";
            vm.ReadFileContentsFromVm( file );
        }

        [Test]
        [ExpectedException( typeof( VirtualMachineIsNotReadyException ) )]
        public void ThrowsAnExceptionWhenTryingToTakeScreenshotOnStoppedVirtualMachine()
        {
            IVirtualMachine vm = CreateVirtualMachine();

            vm.TakeScreenshot( "c:\\screenshot" );
        }
    }
}